use std::{sync::OnceLock, time::Duration};

#[cfg(feature = "web")]
use async_std::task::sleep;
#[cfg(feature = "desktop")]
use dioxus::desktop::{Config, LogicalSize, WindowBuilder};
use dioxus::{logger::tracing, prelude::*};
#[cfg(feature = "bundled")]
use include_dir::{Dir, include_dir};
use models::{Question, Questions};
use rand::seq::SliceRandom;
#[cfg(feature = "desktop")]
use std::{env, fs};
#[cfg(not(feature = "web"))]
use tokio::time::sleep;

mod models;

#[cfg(feature = "bundled")]
const QUESTIONS_DIR: Dir = include_dir!("./questions");
#[cfg(feature = "desktop")]
const DEFAULT_DIR: &str = "./questions";
const MAIN_CSS: Asset = asset!("/assets/main.css");

pub static QUESTIONS: OnceLock<Questions> = OnceLock::new();

fn get_questions() -> Questions {
	QUESTIONS
		.get_or_init(|| {
			let mut questions = Questions::default();

			#[cfg(feature = "bundled")]
			{
				for file in QUESTIONS_DIR.files() {
					if let Some(content) = file.contents_utf8() {
						if let Ok(other_questions) = toml::from_str(content) {
							questions += other_questions;
						}
					}
				}
			}

			#[cfg(feature = "desktop")]
			{
				let args: Vec<String> = env::args().collect();
				if args.len() > 1 {
					for i in 1..args.len() {
						let file_name = args.get(i).map(|x| x.as_str());
						let Some(file_name) = file_name else {
							continue;
						};

						let Ok(questions_str) = std::fs::read_to_string(file_name) else {
							continue;
						};

						let Ok(other_questions) = toml::from_str(&questions_str) else {
							continue;
						};

						questions += other_questions;
					}
				} else if questions.is_empty() {
					if let Ok(dir) = fs::read_dir(DEFAULT_DIR) {
						for file in dir.flatten() {
							if let Ok(questions_str) = std::fs::read_to_string(file.path()) {
								if let Ok(other_questions) = toml::from_str(&questions_str) {
									questions += other_questions;
								}
							}
						}
					}
				}
			}

			questions
		})
		.clone()
}

fn get_rand_questions() -> Vec<Question> {
	let mut questions = get_questions().questions;
	let mut rng = rand::rng();

	// Randomize the answers
	questions
		.iter_mut()
		.for_each(|q| q.answers.shuffle(&mut rng));

	// Randomize the questions list
	questions.shuffle(&mut rng);

	questions
}

fn main() {
	#[cfg(feature = "desktop")]
	{
		dioxus::LaunchBuilder::new()
			.with_cfg(
				Config::default().with_menu(None).with_window(
					WindowBuilder::new()
						.with_min_inner_size(LogicalSize::new(640, 540))
						.with_maximized(true)
						.with_title("Flashcards"),
				),
			)
			.launch(App)
	}

	#[cfg(not(feature = "desktop"))]
	{
		dioxus::launch(App);
	}
}

#[component]
fn App() -> Element {
	let questions = get_rand_questions();
	let mut questions = use_signal(|| questions);

	let mut correct_questions = use_signal(|| 0);
	let mut wrong_questions = use_signal(|| 0);

	let mut has_finished = use_signal(|| false);
	let mut has_restart = use_signal(|| false);

	use_effect(move || {
		let has_restart = has_restart();
		if has_restart {
			let mut questions_lock = questions.write();

			*questions_lock = get_rand_questions();

			correct_questions.set(0);
			wrong_questions.set(0);

			has_finished.set(false);
		}
	});

	use_effect(move || {
		let has_finished = has_finished();
		if !has_finished {
			has_restart.set(false);
		}
	});

	rsx! {
		document::Link { rel: "stylesheet", href: MAIN_CSS }
		if !has_finished() {
			QuestionForm { questions, correct_questions, wrong_questions, has_finished, has_restart }
		}
		else if has_finished() {
			Report {
				questions: questions(),
				correct_questions: correct_questions(),
				wrong_questions: wrong_questions(),
				has_restart,
			}
		}
	}
}

#[allow(clippy::redundant_closure)]
#[component]
pub fn QuestionForm(
	questions: Signal<Vec<Question>>,
	correct_questions: Signal<i32>,
	wrong_questions: Signal<i32>,
	has_finished: Signal<bool>,
	has_restart: Signal<bool>,
) -> Element {
	let mut current = use_signal(move || questions()[0].clone());

	use_effect(move || {
		let _correct_count = correct_questions();
		let _wrong_count = wrong_questions();

		let mut questions_lock = questions.write();

		if questions_lock.is_empty() {
			has_finished.set(true);
		} else {
			let new_question = questions_lock.remove(0);
			current.set(new_question);
		}
	});

	let total_correct = use_memo(move || {
		current()
			.answers
			.iter()
			.filter(|a| a.is_correct.unwrap_or_default())
			.count()
	});
	let actual_correct = use_memo(move || {
		let current_lock = current.read();
		let incorrectly_checked = current_lock
			.answers
			.iter()
			.filter(|a| !a.is_correct.unwrap_or_default() && a.checked)
			.count();
		let correctly_checked = current_lock
			.answers
			.iter()
			.filter(|a| a.is_correct.unwrap_or_default() && a.checked)
			.count();

		correctly_checked.saturating_sub(incorrectly_checked)
	});

	let mut correct_animation = use_signal(|| false);
	let mut wrong_animation = use_signal(|| false);

	use_future(move || async move {
		loop {
			// Sleep will either be from async_std for WASM for tokio for desktop/mobile
			sleep(Duration::from_millis(400)).await;
			let is_correct = correct_animation();
			let is_wrong = wrong_animation();
			if is_correct || is_wrong {
				sleep(Duration::from_millis(600)).await;

				let mut current_lock = current.write();
				let mut correct_questions_lock = correct_questions.write();
				let mut wrong_questions_lock = wrong_questions.write();

				// Uncheck all the answers
				current_lock
					.answers
					.iter_mut()
					.for_each(|a| a.checked = false);

				// Update the correct/wrong counters
				if is_correct {
					*correct_questions_lock += 1;
				} else if is_wrong {
					*wrong_questions_lock += 1;
				}

				// Reset animation flags
				correct_animation.set(false);
				wrong_animation.set(false);
			}
		}
	});

	let answer_buttons = current()
		.answers
		.into_iter()
		.enumerate()
		.map(|(i, _answer)| {
			rsx! {
				AnswerCheckbox { current, id: i }
			}
		});

	rsx! {
		Result { is_correct: correct_animation(), is_wrong: wrong_animation() },
		Info { questions, correct_questions, wrong_questions },
		form {
			id: "form",
			onsubmit: move |_| {
				if actual_correct() == total_correct() {
					correct_animation.set(true);
				}
				else {
					wrong_animation.set(true);
				}
			},

			h1 {
				class: if correct_animation() { "correct" },
				class: if wrong_animation() { "wrong" },
				"{current().message}"
			}
			{ answer_buttons }
			input {
				type: "submit",
				value: "Submit"
			}
		}
	}
}

#[component]
pub fn AnswerCheckbox(current: Signal<Question>, id: usize) -> Element {
	let message = use_memo(move || {
		current
			.read()
			.answers
			.get(id)
			.map(|a| a.message.clone())
			.unwrap_or("???".to_string())
	});

	let checked = use_memo(move || {
		current
			.read()
			.answers
			.get(id)
			.map(|a| a.checked)
			.unwrap_or(false)
	});

	rsx! {
		div {
			class: if checked() { "selected" },
			onclick: move |_| {
				let mut write_lock = current.write();
				if let Some(answer) = write_lock.answers.get_mut(id) {
					answer.checked = !answer.checked;
				}
			},

			input {
				id: "a{id}",
				name: "a{id}",
				type: "checkbox",
				value: "{id}",
				checked: "{checked}",
				onclick: move |event| event.prevent_default(),
			}
			label {
				for: "a{id}",
				onclick: move |event| event.prevent_default(),
				"{message}"
			},
		}
	}
}

#[allow(clippy::redundant_closure)]
#[component]
pub fn Result(is_correct: bool, is_wrong: bool) -> Element {
	let is_visible = is_correct || is_wrong;

	rsx! {
		div {
			id: "result",
			class: if is_visible { "visible" },

			div {
				class: if is_correct { "correct" },
				class: if is_wrong { "wrong" },
				if is_correct { "✓" } else { "X" }
			}
		}
	}
}

#[allow(clippy::redundant_closure)]
#[component]
pub fn Info(
	questions: Signal<Vec<Question>>,
	correct_questions: Signal<i32>,
	wrong_questions: Signal<i32>,
) -> Element {
	let total_questions = get_questions().len();
	let current_question = use_memo(move || questions().len());
	let left_questions = use_memo(move || total_questions - current_question());

	rsx! {
		div {
			id: "counter",

			div { "{left_questions}" }
			div {
				class: "correct",
				"{correct_questions}"
			}
			div {
				class: "wrong",
				"{wrong_questions}"
			}
			div { "{total_questions}" }
		}
	}
}

#[component]
pub fn Report(
	questions: Vec<Question>,
	correct_questions: i32,
	wrong_questions: i32,
	has_restart: Signal<bool>,
) -> Element {
	let questions_size = questions.len();

	rsx! {
		div {
			id: "report",

			p { "Questions: {questions_size}" },
			p {
				class: "correct",
				"Correct: {correct_questions}"
			},
			p {
				class: "wrong",
				"Wrong: {wrong_questions}"
			},
			div {
				button {
					onclick: move |_| {
						has_restart.set(true);
					},
					"Restart"
				},
				button {
					"Restart with only wrong questions"
				}
			}
		}
	}
}
